use super::InputOwned;
use core::{fmt, mem, num::ParseIntError, ops::Range};
use nom::{
    error::{
        ContextError as NomContextError, ErrorKind as NomErrorKind, FromExternalError,
        ParseError as NomParseError,
    },
    InputLength,
};
use std::fmt::Write;
use thiserror::Error;

#[derive(Clone, Debug, PartialEq, Error)]
pub enum ErrorKind {
    #[error("nom error {0:?}")]
    Nom(NomErrorKind),
    #[error("expected '{0}'")]
    ExpectedChar(char),
    #[error("...in {0}")]
    Context(&'static str),
    #[error("invalid integer: {0}")]
    InvalidInteger(#[from] ParseIntError),
    #[error("expected identifier (like /[a-zA-Z][A-Za-z0-9_]*/)")]
    ExpectedIdentifier,
}

#[allow(clippy::len_without_is_empty)]
pub trait ErrorInput: InputLength {
    /// Offset of (start of) other within self.
    /// Returns None if the start of other is not in self,
    /// or if the operation is otherwise not supported
    fn offset(&self, other: &Self) -> Option<usize>;
}

#[allow(clippy::len_without_is_empty)]
pub trait ErrorInputDisplay: fmt::Debug {
    fn len(&self) -> usize;

    fn write_range(&self, range: Range<usize>, f: &mut fmt::Formatter) -> fmt::Result;

    fn pointer() -> &'static str;
    fn line() -> &'static str;
    fn space() -> &'static str;
}

#[derive(Clone, PartialEq, Debug)]
pub struct Error<I> {
    pub input: I,
    pub errors: Vec<(Option<Range<usize>>, ErrorKind)>,
}

impl<I: ErrorInput> Error<I> {
    pub fn new(input: I, kind: ErrorKind) -> Self {
        let errors = vec![(Some(0..input.input_len()), kind)];
        Self { input, errors }
    }

    fn append_kind(&mut self, mut input: I, kind: ErrorKind) {
        let range = if let Some(offset) = self.input.offset(&input) {
            // if the new input is contained in self, just use that
            Some(offset..(offset + input.input_len()))
        } else if let Some(offset) = input.offset(&self.input) {
            // if the new input is before self, replace the self input
            // and update all range offsets
            let input_len = input.input_len();
            mem::swap(&mut self.input, &mut input);
            for err in &mut self.errors {
                if let Some(err_range) = &mut err.0 {
                    err_range.start += offset;
                    err_range.end += offset;
                }
            }
            Some(0..input_len)
        } else {
            None
        };
        self.errors.push((range, kind));
    }
}

#[allow(dead_code)]
pub fn error_context<I: ErrorInput + Clone, F, P, O>(
    mut f: F,
    mut p: P,
) -> impl FnMut(I) -> nom::IResult<I, O, Error<I>>
where
    F: FnMut() -> ErrorKind,
    P: nom::Parser<I, O, Error<I>>,
{
    move |data| {
        p.parse(data.clone()).map_err(|err| {
            err.map(|mut err| {
                err.append_kind(data, f());
                err
            })
        })
    }
}

impl<I: ErrorInput> NomParseError<I> for Error<I> {
    fn from_error_kind(input: I, kind: NomErrorKind) -> Self {
        Self::new(input, ErrorKind::Nom(kind))
    }

    fn append(input: I, kind: NomErrorKind, mut other: Self) -> Self {
        other.append_kind(input, ErrorKind::Nom(kind));
        other
    }

    fn from_char(input: I, c: char) -> Self {
        Self::new(input, ErrorKind::ExpectedChar(c))
    }
}

impl<I: ErrorInput> NomContextError<I> for Error<I> {
    fn add_context(input: I, ctx: &'static str, mut other: Self) -> Self {
        other.append_kind(input, ErrorKind::Context(ctx));
        other
    }
}

impl<I: ErrorInput, E: Into<ErrorKind>> FromExternalError<I, E> for Error<I> {
    fn from_external_error(input: I, _kind: NomErrorKind, e: E) -> Self {
        Self::new(input, e.into())
    }
}

impl<I> Error<I> {
    pub fn map_input<I2, F>(self, mut f: F) -> Error<I2>
    where
        F: FnMut(I) -> I2,
    {
        Error {
            input: f(self.input),
            errors: self.errors,
        }
    }
}

impl<'a> From<Error<&'a str>> for Error<&'a [u8]> {
    fn from(e: Error<&'a str>) -> Self {
        e.map_input(str::as_bytes)
    }
}

impl<I: ToOwned + ?Sized> Error<&I> {
    pub fn into_owned(self) -> Error<I::Owned> {
        self.map_input(I::to_owned)
    }
}

// adapted from https://fasterthanli.me/series/making-our-own-ping/part-9
impl<I: ErrorInputDisplay> fmt::Display for Error<I> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "parsing error")?;

        let margin_left = 4;
        let margin_str = " ".repeat(margin_left);

        // maximum amount of binary data we'll dump per line
        let maxlen = 60;

        // given a big slice, an offset, and a length, attempt to show
        // some data before, some data after, and highlight which part
        // we're talking about with tildes.
        let print_slice = |f: &mut fmt::Formatter, s: &I, range: &Range<usize>| -> fmt::Result {
            // decide which part of `s` we're going to show.
            let (range, offset, len) = {
                // see diagram further in article.
                let offset = range.start;
                let len = range.end - range.start;

                let avail_after = s.len() - offset;
                let after = std::cmp::min(avail_after, maxlen / 2);

                let avail_before = offset;
                let before = std::cmp::min(avail_before, maxlen / 2);

                let new_start = offset - before;
                let new_end = offset + after;
                let new_offset = before;
                let new_len = std::cmp::min(new_end - new_start, len);

                (new_start..new_end, new_offset, new_len)
            };

            f.write_str(&margin_str)?;
            s.write_range(range.clone(), f)?;
            writeln!(f)?;

            write!(f, "{}", margin_str)?;
            for i in 0..range.len() {
                // each byte takes three characters, ie "FF "
                if i == offset + len - 1 {
                    // ..except the last one
                    f.write_str(I::pointer())?;
                } else if (offset..offset + len).contains(&i) {
                    f.write_str(I::line())?;
                } else {
                    f.write_str(I::space())?;
                };
            }
            writeln!(f)?;

            Ok(())
        };

        for (range, kind) in self.errors.iter().rev() {
            writeln!(f, "{}", kind)?;
            if let Some(range) = range {
                print_slice(f, &self.input, range)?;
            }
        }
        Ok(())
    }
}

// // TODO: should this really be the same as Display?
// impl<I: ErrorInputDisplay> fmt::Debug for Error<I> {
//     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
//         <Self as fmt::Display>::fmt(self, f)
//     }
// }

impl<I: ErrorInputDisplay> std::error::Error for Error<I> {}

impl<'a> ErrorInput for &'a [u8] {
    fn offset(&self, other: &Self) -> Option<usize> {
        (other.as_ptr() as usize)
            .checked_sub(self.as_ptr() as usize)
            .filter(|&offset| offset < self.len())
    }
}

impl<'a> ErrorInputDisplay for &'a [u8] {
    fn len(&self) -> usize {
        <[u8]>::len(self)
    }
    fn write_range(&self, range: std::ops::Range<usize>, f: &mut fmt::Formatter) -> fmt::Result {
        for b in &self[range] {
            write!(f, "{:02X} ", b)?;
        }
        Ok(())
    }
    fn pointer() -> &'static str {
        "~~"
    }
    fn line() -> &'static str {
        "~~~"
    }
    fn space() -> &'static str {
        "   "
    }
}

impl ErrorInputDisplay for InputOwned {
    fn len(&self) -> usize {
        Vec::len(self)
    }
    fn write_range(&self, range: std::ops::Range<usize>, f: &mut fmt::Formatter) -> fmt::Result {
        self.as_slice().write_range(range, f)
    }
    fn pointer() -> &'static str {
        <&'static [u8] as ErrorInputDisplay>::pointer()
    }
    fn line() -> &'static str {
        <&'static [u8] as ErrorInputDisplay>::line()
    }
    fn space() -> &'static str {
        <&'static [u8] as ErrorInputDisplay>::space()
    }
}

impl<'a> ErrorInput for &'a str {
    fn offset(&self, other: &Self) -> Option<usize> {
        (other.as_ptr() as usize).checked_sub(self.as_ptr() as usize)
    }
}

impl<'a> ErrorInputDisplay for &'a str {
    fn len(&self) -> usize {
        str::len(self)
    }
    fn write_range(&self, range: std::ops::Range<usize>, f: &mut fmt::Formatter) -> fmt::Result {
        for c in self[range].chars() {
            f.write_char(match c {
                '\r' => '␍',
                '\n' => '␊',
                c if c.is_control() => '␣',
                c => c,
            })?;
        }
        Ok(())
    }
    fn pointer() -> &'static str {
        "~"
    }
    fn line() -> &'static str {
        "~"
    }
    fn space() -> &'static str {
        " "
    }
}

impl ErrorInputDisplay for String {
    fn len(&self) -> usize {
        String::len(self)
    }
    fn write_range(&self, range: std::ops::Range<usize>, f: &mut fmt::Formatter) -> fmt::Result {
        self.as_str().write_range(range, f)
    }
    fn pointer() -> &'static str {
        <&'static str as ErrorInputDisplay>::pointer()
    }
    fn line() -> &'static str {
        <&'static str as ErrorInputDisplay>::line()
    }
    fn space() -> &'static str {
        <&'static str as ErrorInputDisplay>::space()
    }
}
