use std::io::SeekFrom;

use nom::number::streaming::le_u32;

use crate::chunk::ChunkHeader;
use crate::error::{Error, Result};
use crate::index::{ChunkInfo, ConnInfo};
use crate::parse::{self, Header, Op, Version};

mod io;
#[cfg(feature = "mmap")]
mod mmap;

pub use io::IoReader;
#[cfg(feature = "mmap")]
pub use mmap::MmapReader;

pub trait BagReader {
    fn read_parser<'a, O: 'a, P>(&'a mut self, parser: P) -> Result<O>
    where
        P: nom::Parser<&'a [u8], O, parse::Error<&'a [u8]>>;

    fn seek(&mut self, pos: SeekFrom) -> Result<()>;

    fn read_version(&mut self) -> Result<Version> {
        self.read_parser(Version::parse)
    }

    fn read_header(&mut self) -> Result<Header> {
        self.read_parser(Header::parse)
    }

    fn read_header_op(&mut self, op: Op) -> Result<Header> {
        let header = self.read_header()?;
        if header.read_op()? != op {
            return Err(Error::other(format!("Expected {:?} op", op)));
        }
        Ok(header)
    }

    fn read_data_length(&mut self) -> Result<u32> {
        self.read_parser(le_u32)
    }

    fn skip_data(&mut self) -> Result<()> {
        let data_length = self.read_data_length()?;
        self.seek(SeekFrom::Current(data_length as i64))?;
        Ok(())
    }

    fn read_conn_info(&mut self) -> Result<ConnInfo> {
        let record_header = self.read_header_op(Op::Connection)?;
        let conn_header = self.read_header()?;
        ConnInfo::from_headers(record_header, conn_header)
    }

    fn read_chunk_info(&mut self) -> Result<ChunkInfo> {
        let header = self.read_header_op(Op::ChunkInfo)?;
        // TODO: read connection message counts
        self.skip_data()?;
        ChunkInfo::from_header(header)
    }

    fn read_chunk_header(&mut self, pos: u64) -> Result<ChunkHeader> {
        self.seek(SeekFrom::Start(pos))?;
        let header = self.read_header_op(Op::Chunk)?;
        ChunkHeader::from_header(header)
    }
}
